from typing import Any

import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import ValueValidationError


@pytest.mark.parametrize("value", [
    None,
    True,
    42,
    3.14,
    "banana",
])
def test_const_validation(value: Any):
    with when:
        result = validate(schema.const(value), value)

    with then:
        assert result.get_errors() == []


def test_const_validation_error():
    with given:
        expected_value = "banana"
        actual_value = "apple"

    with when:
        result = validate(schema.const(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value)
        ]


def test_const_type_validation_kwargs():
    with given:
        expected_value = "banana"
        actual_value = "apple"
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.const(expected_value), actual_value, path=path)

    with then:
        assert result.get_errors() == [
            ValueValidationError(path, actual_value, expected_value)
        ]



import uuid

from baby_steps import given, then, when
from district42.types import IntSchema
from th import PathHolder

from valera.errors import (
    AlphabetValidationError,
    ExtraElementValidationError,
    ExtraKeyValidationError,
    InvalidUUIDVersionValidationError,
    LengthValidationError,
    MaxLengthValidationError,
    MaxValueValidationError,
    MinLengthValidationError,
    MinValueValidationError,
    MissingElementValidationError,
    MissingKeyValidationError,
    RegexValidationError,
    SchemaMismatchValidationError,
    SubstrValidationError,
    TypeValidationError,
    ValueValidationError,
)


def test_validation_type_error():
    with when:
        res = TypeValidationError(PathHolder(), "banana", int)

    with then:
        assert repr(res) == "TypeValidationError(PathHolder(), 'banana', <class 'int'>)"


def test_validation_value_error():
    with given:
        actual = "banana"
        expected = "cucumber"

    with when:
        res = ValueValidationError(PathHolder(), actual, expected)

    with then:
        assert repr(res) == "ValueValidationError(PathHolder(), 'banana', 'cucumber')"


def test_validation_min_value_error():
    with when:
        res = MinValueValidationError(PathHolder(), 41, 42)

    with then:
        assert repr(res) == "MinValueValidationError(PathHolder(), 41, 42)"


def test_validation_max_value_error():
    with when:
        res = MaxValueValidationError(PathHolder(), 43, 42)

    with then:
        assert repr(res) == "MaxValueValidationError(PathHolder(), 43, 42)"


def test_validation_len_error():
    with when:
        res = LengthValidationError(PathHolder(), "banana", 7)

    with then:
        assert repr(res) == "LengthValidationError(PathHolder(), 'banana', 7)"


def test_validation_min_len_error():
    with when:
        res = MinLengthValidationError(PathHolder(), "banana", 7)

    with then:
        assert repr(res) == "MinLengthValidationError(PathHolder(), 'banana', 7)"


def test_validation_max_len_error():
    with when:
        res = MaxLengthValidationError(PathHolder(), "banana", 7)

    with then:
        assert repr(res) == "MaxLengthValidationError(PathHolder(), 'banana', 7)"


def test_validation_alphabet_error():
    with when:
        res = AlphabetValidationError(PathHolder(), "banana!", "abn")

    with then:
        assert repr(res) == "AlphabetValidationError(PathHolder(), 'banana!', 'abn')"


def test_validation_substr_error():
    with when:
        res = SubstrValidationError(PathHolder(), "value", "substr")

    with then:
        assert repr(res) == "SubstrValidationError(PathHolder(), 'value', 'substr')"


def test_validation_regex_error():
    with when:
        res = RegexValidationError(PathHolder(), "value", ".*")

    with then:
        assert repr(res) == "RegexValidationError(PathHolder(), 'value', '.*')"


def test_validation_index_error():
    with when:
        res = MissingElementValidationError(PathHolder(), [], 0)

    with then:
        assert repr(res) == "MissingElementValidationError(PathHolder(), [], 0)"


def test_validation_extra_element_error():
    with when:
        res = ExtraElementValidationError(PathHolder(), [1], 0)

    with then:
        assert repr(res) == "ExtraElementValidationError(PathHolder(), [1], 0)"


def test_validation_missing__key_error():
    with when:
        res = MissingKeyValidationError(PathHolder(), {}, "key")

    with then:
        assert repr(res) == "MissingKeyValidationError(PathHolder(), {}, 'key')"


def test_validation_extra_key_error():
    with when:
        res = ExtraKeyValidationError(PathHolder(), {"key": "1"}, "key")

    with then:
        assert repr(res) == "ExtraKeyValidationError(PathHolder(), {'key': '1'}, 'key')"


def test_validation_schema_mismatch_error():
    with when:
        res = SchemaMismatchValidationError(PathHolder(), "key", (IntSchema(),))

    with then:
        assert repr(res) == "SchemaMismatchValidationError(PathHolder(), 'key', (schema.int,))"


def test_validation_invalid_uuid_version_error():
    with given:
        uuid5 = uuid.UUID("886313e1-3b8a-5372-9b90-0c9aee199e5d", version=5)

    with when:
        res = InvalidUUIDVersionValidationError(PathHolder(), uuid5, 5, 4)

    with then:
        assert repr(res) == ("InvalidUUIDVersionValidationError(PathHolder(), "
                             "UUID('886313e1-3b8a-5372-9b90-0c9aee199e5d'), 5, 4)")



from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import TypeValidationError


def test_none_validation():
    with when:
        result = validate(schema.none, None)

    with then:
        assert result.get_errors() == []


def test_none_validation_error():
    with given:
        value = False

    with when:
        result = validate(schema.none, value)

    with then:
        assert result.get_errors() == [
            TypeValidationError(PathHolder(), value, type(None))
        ]


def test_none_type_validation_kwargs():
    with given:
        actual_value = False
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.none, actual_value, path=path)

    with then:
        assert result.get_errors() == [
            TypeValidationError(path, actual_value, type(None))
        ]



from baby_steps import then, when
from district42 import schema
from pytest import raises

from valera import ValidationException, ValidationResult, eq, validate, validate_or_fail


def test_validator_validate_pass():
    with when:
        result = validate(schema.int, 42)

    with then:
        assert isinstance(result, ValidationResult)
        assert not result.has_errors()


def test_validator_validate_fail():
    with when:
        result = validate(schema.int, "42")

    with then:
        assert isinstance(result, ValidationResult)
        assert result.has_errors()


def test_validator_validate_or_fail():
    with when:
        result = validate_or_fail(schema.int, 42)

    with then:
        assert result is True


def test_validator_validate_or_fail_error():
    with when, raises(Exception) as exception:
        validate_or_fail(schema.dict, "")

    with then:
        assert exception.type is ValidationException
        assert str(exception.value) == ("\n - Value '' must be <class 'dict'>, "
                                        "but <class 'str'> given")


def test_validator_validate_or_fail_multiline():
    with when, raises(Exception) as exception:
        validate_or_fail(schema.dict({"id": schema.int, "name": schema.str}), {})

    with then:
        assert exception.type is ValidationException
        assert str(exception.value) == ("\n - Key _['id'] does not exist"
                                        "\n - Key _['name'] does not exist")


def test_validator_eq():
    with when:
        result = eq(schema.str, schema.str)

    with then:
        assert result is True


def test_validator_not_eq_schema():
    with when:
        result = eq(schema.str, schema.int)

    with then:
        assert result is False


def test_validator_not_eq_props():
    with when:
        result = eq(schema.str, schema.str.len(1))

    with then:
        assert result is False


def test_validator_not_eq_value():
    with when:
        result = eq(schema.str, 42)

    with then:
        assert result is False



from baby_steps import given, then, when
from pytest import raises

from valera import Formatter
from valera.errors import ValidationError


def test_register_formatter():
    with given:
        formatted = "<custom_error>"

        class CustomError(ValidationError):
            def format(self, formatter: Formatter) -> str:
                return formatter.format_custom_error(self)

        class CustomFormatter(Formatter, extend=True):
            def format_custom_error(self, error: CustomError) -> str:
                return formatted

    with when:
        res = CustomError().format(CustomFormatter())

    with then:
        assert res == formatted


def test_register_formatter_without_extend():
    with given:
        formatted = "<another_custom_error>"

        class AnotherCustomError(ValidationError):
            def format(self, formatter: Formatter) -> str:
                return formatter.format_another_custom_error(self)

        class AnotherCustomFormatter(Formatter):
            def format_custom_error(self, error: AnotherCustomError) -> str:
                return formatted

    with when, raises(Exception) as exception:
        AnotherCustomError().format(AnotherCustomFormatter())

    with then:
        assert exception.type is AttributeError



import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import (
    MaxValueValidationError,
    MinValueValidationError,
    TypeValidationError,
    ValueValidationError,
)


def test_int_type_validation():
    with when:
        result = validate(schema.int, 42)

    with then:
        assert result.get_errors() == []


def test_int_type_validation_error():
    with given:
        value = 3.14

    with when:
        result = validate(schema.int, value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder(), value, int)]


def test_int_value_validation():
    with given:
        value = 42

    with when:
        result = validate(schema.int(value), value)

    with then:
        assert result.get_errors() == []


def test_int_value_validation_error():
    with given:
        expected_value = 42
        actual_value = 43

    with when:
        result = validate(schema.int(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value)
        ]


@pytest.mark.parametrize("value", [1, 2])
def test_int_min_value_validation(value: int):
    with when:
        result = validate(schema.int.min(1), value)

    with then:
        assert result.get_errors() == []


def test_int_min_value_validation_error():
    with given:
        min_value = 1
        actual_value = 0

    with when:
        result = validate(schema.int.min(min_value), actual_value)

    with then:
        assert result.get_errors() == [
            MinValueValidationError(PathHolder(), actual_value, min_value)
        ]


@pytest.mark.parametrize("value", [1, 2])
def test_int_max_value_validation(value: int):
    with when:
        result = validate(schema.int.max(2), value)

    with then:
        assert result.get_errors() == []


def test_int_max_value_validation_error():
    with given:
        max_value = 1
        actual_value = 2

    with when:
        result = validate(schema.int.max(max_value), actual_value)

    with then:
        assert result.get_errors() == [
            MaxValueValidationError(PathHolder(), actual_value, max_value)
        ]


@pytest.mark.parametrize("value", [1, 2, 3])
def test_int_min_max_value_validation(value: int):
    with when:
        result = validate(schema.int.min(1).max(3), value)

    with then:
        assert result.get_errors() == []


def test_int_min_max_greater_value_validation_error():
    with given:
        min_value = 1
        max_value = 3
        actual_value = 4

    with when:
        result = validate(schema.int.min(min_value).max(max_value), actual_value)

    with then:
        assert result.get_errors() == [
            MaxValueValidationError(PathHolder(), actual_value, max_value)
        ]


def test_int_min_max_less_value_validation_error():
    with given:
        min_value = 1
        max_value = 3
        actual_value = 0

    with when:
        result = validate(schema.int.min(min_value).max(max_value), actual_value)

    with then:
        assert result.get_errors() == [
            MinValueValidationError(PathHolder(), actual_value, min_value)
        ]


def test_int_type_validation_kwargs():
    with given:
        expected_value = 42
        actual_value = 43
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.int(expected_value), actual_value, path=path)

    with then:
        assert result.get_errors() == [
            ValueValidationError(path, actual_value, expected_value)
        ]



from copy import deepcopy
from typing import Any, Sequence

from th import PathHolder

from ._abstract_formatter import AbstractFormatter
from .errors import (
    AlphabetValidationError,
    ExtraElementValidationError,
    ExtraKeyValidationError,
    InvalidUUIDVersionValidationError,
    LengthValidationError,
    MaxLengthValidationError,
    MaxValueValidationError,
    MinLengthValidationError,
    MinValueValidationError,
    MissingElementValidationError,
    MissingKeyValidationError,
    RegexValidationError,
    SchemaMismatchValidationError,
    SubstrValidationError,
    TypeValidationError,
    ValueValidationError,
)

__all__ = ("Formatter",)


class Formatter(AbstractFormatter):
    def __init__(self, root: str = "_") -> None:
        self._root = root

    @property
    def root(self) -> str:
        return self._root

    def _format_path(self, path: PathHolder) -> str:
        return str(path.__class__(self._root, [x for x in path]))

    def _at_path(self, path: PathHolder) -> str:
        return " at " + self._format_path(path) if len(path) > 0 else ""

    def _get_type(self, value: Any) -> str:
        return str(type(value))

    def _pluralize(self, count: int, options: Sequence[str]) -> str:
        return options[0] if count == 1 else options[-1]

    def format_type_error(self, error: TypeValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {error.actual_value!r}{formatted_path} "
                f"must be {error.expected_type}, but {actual_type} given")

    def format_value_error(self, error: ValueValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {actual_type}{formatted_path} "
                f"must be equal to {error.expected_value!r}, but {error.actual_value!r} given")

    def format_min_value_error(self, error: MinValueValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {actual_type}{formatted_path} "
                f"must be greater than or equal to {error.min_value!r}, "
                f"but {error.actual_value!r} given")

    def format_max_value_error(self, error: MaxValueValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {actual_type}{formatted_path} "
                f"must be less than or equal to {error.max_value!r}, "
                f"but {error.actual_value!r} given")

    def _format_length_error(self, path: PathHolder, actual_value: Any,
                             length: int, accuracy: str) -> str:
        actual_type = self._get_type(actual_value)
        actual_len = len(actual_value)
        formatted_path = self._at_path(path)
        options = ("element", "elements")
        return (
            f"Value {actual_type}{formatted_path} "
            f"must have {accuracy} {length} {self._pluralize(length, options)}, "
            f"but it has {actual_len} {self._pluralize(actual_len, options)}"
        )

    def format_length_error(self, error: LengthValidationError) -> str:
        return self._format_length_error(error.path, error.actual_value,
                                         error.length, "exactly")

    def format_min_length_error(self, error: MinLengthValidationError) -> str:
        return self._format_length_error(error.path, error.actual_value,
                                         error.min_length, "at least")

    def format_max_length_error(self, error: MaxLengthValidationError) -> str:
        return self._format_length_error(error.path, error.actual_value,
                                         error.max_length, "at most")

    def format_alphabet_error(self, error: AlphabetValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {actual_type}{formatted_path} "
                f"must contain only {error.alphabet!r}, but {error.actual_value!r} given")

    def format_substr_error(self, error: SubstrValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {actual_type}{formatted_path} "
                f"must contain {error.substr!r}, but {error.actual_value!r} given")

    def format_regex_error(self, error: RegexValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {actual_type}{formatted_path} "
                f"must match pattern {error.pattern!r}, but {error.actual_value!r} given")

    def format_missing_element_error(self, error: MissingElementValidationError) -> str:
        path = deepcopy(error.path)
        formatted_path = self._format_path(path[error.index])
        return f"Element {formatted_path} does not exist"

    def format_extra_element_error(self, error: ExtraElementValidationError) -> str:
        formatted_path = self._at_path(error.path)
        return f"Value{formatted_path} contains extra element at index {error.index!r}"

    def format_missing_key_error(self, error: MissingKeyValidationError) -> str:
        path = deepcopy(error.path)
        formatted_path = self._format_path(path[error.missing_key])
        return f"Key {formatted_path} does not exist"

    def format_extra_key_error(self, error: ExtraKeyValidationError) -> str:
        formatted_path = self._at_path(error.path)
        return f"Value{formatted_path} contains extra key {error.extra_key!r}"

    def format_schema_missmatch_error(self, error: SchemaMismatchValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {actual_type}{formatted_path} "
                f"must match any of {error.expected_schemas!r}, but {error.actual_value!r} given")

    def format_invalid_uuid_version_error(self, error: InvalidUUIDVersionValidationError) -> str:
        actual_type = self._get_type(error.actual_value)
        formatted_path = self._at_path(error.path)
        return (f"Value {actual_type}{formatted_path} "
                f"must be a UUID version {error.expected_version!r}, "
                f"but {error.actual_value!r} version {error.actual_version!r} given")



from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import TypeValidationError, ValueValidationError


def test_bytes_type_validation():
    with when:
        result = validate(schema.bytes, b"banana")

    with then:
        assert result.get_errors() == []


def test_bytes_type_validation_error():
    with given:
        value = "banana"

    with when:
        result = validate(schema.bytes, value)

    with then:
        assert result.get_errors() == [
            TypeValidationError(PathHolder(), value, bytes),
        ]


def test_bytes_value_validation():
    with given:
        value = b"banana"

    with when:
        result = validate(schema.bytes(value), value)

    with then:
        assert result.get_errors() == []


def test_bytes_value_validation_error():
    with given:
        expected_value = b"banana"
        actual_value = b"cucumber"

    with when:
        result = validate(schema.bytes(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value),
        ]



from baby_steps import given, then, when
from district42 import schema
from district42.types import TypeAliasSchema
from th import PathHolder

from valera import validate
from valera.errors import MinValueValidationError


def test_alias_default_validation():
    with given:
        sch = TypeAliasSchema()

    with when:
        result = validate(sch, None)

    with then:
        assert result.get_errors() == []


def test_alias_validation():
    with given:
        sch = schema.alias("uint", schema.int.min(0))

    with when:
        result = validate(sch, 42)

    with then:
        assert result.get_errors() == []


def test_alias_validation_error():
    with given:
        min_value = 0
        sch = schema.alias("uint", schema.int.min(min_value))
        actual_value = -1

    with when:
        result = validate(sch, actual_value)

    with then:
        assert result.get_errors() == [
            MinValueValidationError(PathHolder(), actual_value, min_value)
        ]



import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import (
    AlphabetValidationError,
    LengthValidationError,
    MaxLengthValidationError,
    MinLengthValidationError,
    RegexValidationError,
    SubstrValidationError,
    TypeValidationError,
    ValueValidationError,
)


def test_str_type_validation():
    with when:
        result = validate(schema.str, "banana")

    with then:
        assert result.get_errors() == []


def test_str_type_validation_error():
    with given:
        value = ["b", "a", "n", "a", "n", "a"]

    with when:
        result = validate(schema.str, value)

    with then:
        assert result.get_errors() == [
            TypeValidationError(PathHolder(), value, str),
        ]


def test_str_value_validation():
    with given:
        value = "banana"

    with when:
        result = validate(schema.str(value), value)

    with then:
        assert result.get_errors() == []


def test_str_value_validation_error():
    with given:
        expected_value = "banana"
        actual_value = "cucumber"

    with when:
        result = validate(schema.str(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value),
        ]


def test_str_len_validation():
    with given:
        value = "banana"

    with when:
        result = validate(schema.str.len(6), value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize("value", [
    "banana",
    "banana!!"
])
def test_str_len_validation_error(value: str):
    with given:
        length = 7

    with when:
        result = validate(schema.str.len(length), value)

    with then:
        assert result.get_errors() == [
            LengthValidationError(PathHolder(), value, length),
        ]


@pytest.mark.parametrize("value", [
    "banana",
    "banana!",
])
def test_str_min_len_validation(value: str):
    with when:
        result = validate(schema.str.len(6, ...), value)

    with then:
        assert result.get_errors() == []


def test_str_min_len_validation_error():
    with given:
        value = "banana"
        min_length = 7

    with when:
        result = validate(schema.str.len(min_length, ...), value)

    with then:
        assert result.get_errors() == [
            MinLengthValidationError(PathHolder(), value, min_length),
        ]


@pytest.mark.parametrize("value", [
    "banana",
    "banana!",
])
def test_str_max_len_validation(value: str):
    with when:
        result = validate(schema.str.len(..., 7), value)

    with then:
        assert result.get_errors() == []


def test_str_max_len_validation_error():
    with given:
        value = "banana"
        max_length = 5

    with when:
        result = validate(schema.str.len(..., max_length), value)

    with then:
        assert result.get_errors() == [
            MaxLengthValidationError(PathHolder(), value, max_length)
        ]


@pytest.mark.parametrize(("min_length", "max_length"), [
    (6, 6),
    (5, 7),
])
def test_str_min_max_len_validation(min_length: int, max_length: int):
    with given:
        value = "banana"

    with when:
        result = validate(schema.str.len(min_length, max_length), value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize(("min_length", "max_length"), [
    (5, 5),
    (7, 7),
])
def test_str_min_max_len_validation_error(min_length: int, max_length: int):
    with given:
        value = "banana"

    with when:
        result = validate(schema.str.len(min_length, max_length), value)

    with then:
        assert len(result.get_errors()) == 1


@pytest.mark.parametrize("value", [
    "",
    "11",
    "1234567890",
])
def test_str_alphabet_validation(value: str):
    with when:
        result = validate(schema.str.alphabet("1234567890"), value)

    with then:
        assert result.get_errors() == []


def test_str_alphabet_validation_error():
    with given:
        value = "banana"
        alphabet = "1234567890"

    with when:
        result = validate(schema.str.alphabet(alphabet), value)

    with then:
        assert result.get_errors() == [
            AlphabetValidationError(PathHolder(), value, alphabet)
        ]


def test_str_len_alphabet_validation_error():
    with given:
        value = "banana"
        length = 7
        alphabet = "1234567890"

    with when:
        result = validate(schema.str.alphabet(alphabet).len(length), value)

    with then:
        assert result.get_errors() == [
            LengthValidationError(PathHolder(), value, length),
            AlphabetValidationError(PathHolder(), value, alphabet),
        ]


@pytest.mark.parametrize("value", [
    "banana",
    " banana",
    "banana ",
    " banana ",
])
def test_str_contains_validation(value: str):
    with given:
        substr = "banana"

    with when:
        result = validate(schema.str.contains(substr), value)

    with then:
        assert result.get_errors() == []


def test_str_contains_empty_validation():
    with given:
        value = "banana"
        substr = ""

    with when:
        result = validate(schema.str.contains(substr), value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize("value", [
    "",
    "ananab",
])
def test_str_contains_validation_error(value: str):
    with given:
        substr = "banana"

    with when:
        result = validate(schema.str.contains(substr), value)

    with then:
        assert result.get_errors() == [
            SubstrValidationError(PathHolder(), value, substr)
        ]


def test_str_regex_validation():
    with given:
        pattern = "[a-z]+"
        value = "banana"

    with when:
        result = validate(schema.str.regex(pattern), value)

    with then:
        assert result.get_errors() == []


def test_str_regex_validation_error():
    with given:
        pattern = "[0-9]+"
        value = "banana"

    with when:
        result = validate(schema.str.regex(pattern), value)

    with then:
        assert result.get_errors() == [
            RegexValidationError(PathHolder(), value, pattern)
        ]


def test_str_type_validation_kwargs():
    with given:
        expected_value = "banana"
        actual_value = "cucumber"
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.str(expected_value), actual_value, path=path)

    with then:
        assert result.get_errors() == [
            ValueValidationError(path, actual_value, expected_value)
        ]



from typing import List, Optional

from .errors import ValidationError

__all__ = ("ValidationResult",)


class ValidationResult:
    def __init__(self, errors: Optional[List[ValidationError]] = None) -> None:
        self._errors = errors if (errors is not None) else []

    def add_error(self, error: ValidationError) -> "ValidationResult":
        self._errors.append(error)
        return self

    def add_errors(self, errors: List[ValidationError]) -> "ValidationResult":
        for error in errors:
            self._errors.append(error)
        return self

    def has_errors(self) -> bool:
        return len(self._errors) > 0

    def get_errors(self) -> List[ValidationError]:
        return self._errors

    def __repr__(self) -> str:
        errors = repr(self._errors) if self.has_errors() else ""
        return f"{self.__class__.__name__}({errors})"



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Tuple, Type

from district42 import GenericSchema
from th import PathHolder

if TYPE_CHECKING:
    from .._formatter import Formatter

__all__ = ("ValidationError", "TypeValidationError", "ValueValidationError",
           "MinValueValidationError", "MaxValueValidationError", "LengthValidationError",
           "MinLengthValidationError", "MaxLengthValidationError", "AlphabetValidationError",
           "SubstrValidationError", "RegexValidationError", "MissingElementValidationError",
           "ExtraElementValidationError", "MissingKeyValidationError", "ExtraKeyValidationError",
           "SchemaMismatchValidationError", "InvalidUUIDVersionValidationError",)


class ValidationError(ABC):
    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)

    @abstractmethod
    def format(self, formatter: "Formatter") -> str:
        pass


class TypeValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, expected_type: Type[Any]) -> None:
        self.path = path
        self.actual_value = actual_value
        self.expected_type = expected_type

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_type_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.expected_type!r})")


class ValueValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, expected_value: Any) -> None:
        self.path = path
        self.actual_value = actual_value
        self.expected_value = expected_value

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_value_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.expected_value!r})")


class MinValueValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, min_value: Any) -> None:
        self.path = path
        self.actual_value = actual_value
        self.min_value = min_value

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_min_value_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.min_value!r})")


class MaxValueValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, max_value: Any) -> None:
        self.path = path
        self.actual_value = actual_value
        self.max_value = max_value

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_max_value_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.max_value!r})")


class LengthValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, length: int) -> None:
        self.path = path
        self.actual_value = actual_value
        self.length = length

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_length_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.length!r})")


class MinLengthValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, min_length: int) -> None:
        self.path = path
        self.actual_value = actual_value
        self.min_length = min_length

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_min_length_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.min_length!r})")


class MaxLengthValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, max_length: int) -> None:
        self.path = path
        self.actual_value = actual_value
        self.max_length = max_length

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_max_length_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.max_length!r})")


class AlphabetValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: str, alphabet: str) -> None:
        self.path = path
        self.actual_value = actual_value
        self.alphabet = alphabet

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_alphabet_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.alphabet!r})")


class SubstrValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, substr: str) -> None:
        self.path = path
        self.actual_value = actual_value
        self.substr = substr

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_substr_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.substr!r})")


class RegexValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, pattern: str) -> None:
        self.path = path
        self.actual_value = actual_value
        self.pattern = pattern

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_regex_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.pattern!r})")


class MissingElementValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, index: int) -> None:
        self.path = path
        self.actual_value = actual_value
        self.index = index

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_missing_element_error(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, {self.index!r})"


class ExtraElementValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, index: int) -> None:
        self.path = path
        self.actual_value = actual_value
        self.index = index

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_extra_element_error(self)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, {self.index!r})"


class MissingKeyValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, missing_key: Any) -> None:
        self.path = path
        self.actual_value = actual_value
        self.missing_key = missing_key

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_missing_key_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.missing_key!r})")


class ExtraKeyValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any, extra_key: Any) -> None:
        self.path = path
        self.actual_value = actual_value
        self.extra_key = extra_key

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_extra_key_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.extra_key!r})")


class SchemaMismatchValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any,
                 expected_schemas: Tuple[GenericSchema, ...]) -> None:
        self.path = path
        self.actual_value = actual_value
        self.expected_schemas = expected_schemas

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_schema_missmatch_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.expected_schemas!r})")


class InvalidUUIDVersionValidationError(ValidationError):
    def __init__(self, path: PathHolder, actual_value: Any,
                 actual_version: int, expected_version: int) -> None:
        self.path = path
        self.actual_value = actual_value
        self.actual_version = actual_version
        self.expected_version = expected_version

    def format(self, formatter: "Formatter") -> str:
        return formatter.format_invalid_uuid_version_error(self)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}({self.path!r}, {self.actual_value!r}, "
                f"{self.actual_version!r}, {self.expected_version!r})")



from typing import Any, Dict

import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import ExtraKeyValidationError, MissingKeyValidationError, TypeValidationError


@pytest.mark.parametrize("value", [
    {},
    {"id": 1},
    {"id": 1, "name": "Bob"},
])
def test_dict_type_validation(value: Dict[Any, Any]):
    with given:
        sch = schema.dict

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_dict_type_validation_error():
    with given:
        sch = schema.dict
        value = []

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder(), value, dict)]


def test_dict_no_keys_validation():
    with given:
        sch = schema.dict({})
        value = {}

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_dict_keys_validation():
    with given:
        sch = schema.dict({
            "id": schema.int,
            "name": schema.str,
        })
        value = {"id": 1, "name": "Bob"}

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_dict_extra_key_validation_error():
    with given:
        sch = schema.dict({})
        value = {"id": 1}

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ExtraKeyValidationError(PathHolder(), value, "id")
        ]


def test_dict_missing_key_validation_error():
    with given:
        sch = schema.dict({
            "id": schema.int,
            "name": schema.str,
        })
        value = {"id": 1}

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            MissingKeyValidationError(PathHolder(), value, "name")
        ]


@pytest.mark.parametrize("value", [
    {},
    {"id": 1},
    {"id": 1, "name": "Bob"},
])
def test_dict_relaxed_no_keys_validation(value: Dict[Any, Any]):
    with given:
        sch = schema.dict({...: ...})

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_dict_relaxed_keys_validation():
    with given:
        sch = schema.dict({
            "id": schema.int,
            "name": schema.str,
            ...: ...,
        })
        value = {"id": 1, "name": "Bob", "is_deleted": False}

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_dict_nested_keys_validation():
    with given:
        sch = schema.dict({
            "result": schema.dict({
                "id": schema.int,
                "name": schema.str,
            })
        })
        value = {
            "result": {
                "id": 1,
                "name": "Bob"
            }
        }

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_dict_nested_extra_key_valudidation_error():
    with given:
        sch = schema.dict({
            "result": schema.dict({
                "id": schema.int,
            })
        })
        value = {
            "result": {
                "id": 1,
                "name": "Bob",
            }
        }

    with when:
        result = validate(sch, value)

    with then:
        path = PathHolder()["result"]
        assert result.get_errors() == [
            ExtraKeyValidationError(path, value["result"], "name")
        ]


def test_dict_nested_missing_key_valudidation_error():
    with given:
        sch = schema.dict({
            "result": schema.dict({
                "id": schema.int,
                "name": schema.str,
            })
        })
        value = {
            "result": {
                "id": 1
            }
        }

    with when:
        result = validate(sch, value)

    with then:
        path = PathHolder()["result"]
        assert result.get_errors() == [
            MissingKeyValidationError(path, value["result"], "name")
        ]


def test_dict_validation_kwargs():
    with given:
        sch = schema.dict({
            "id": schema.int,
            "name": schema.str,
        })
        value = {"id": 1}
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(sch, value, path=path)

    with then:
        assert result.get_errors() == [
            MissingKeyValidationError(path, value, "name")
        ]



from typing import Any

import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import SchemaMismatchValidationError


@pytest.mark.parametrize("value", [
    None,
    True,
    42,
    3.14,
    "banana",
    [],
    {},
    Exception,
])
def test_any_validation(value: Any):
    with when:
        result = validate(schema.any, value)

    with then:
        assert result.get_errors() == []


def test_any_type_validation():
    with when:
        result = validate(schema.any(schema.none), None)

    with then:
        assert result.get_errors() == []


def test_any_type_validation_error():
    with given:
        value = False
        types = (schema.none,)

    with when:
        result = validate(schema.any(*types), value)

    with then:
        assert result.get_errors() == [
            SchemaMismatchValidationError(PathHolder(), value, types)
        ]


@pytest.mark.parametrize("value", [42, "42"])
def test_any_types_validation(value: Any):
    with when:
        result = validate(schema.any(schema.int, schema.str), value)

    with then:
        assert result.get_errors() == []


def test_any_types_validation_error():
    with given:
        value = None
        types = (schema.int, schema.str)

    with when:
        result = validate(schema.any(*types), value)

    with then:
        assert result.get_errors() == [
            SchemaMismatchValidationError(PathHolder(), value, types)
        ]


def test_any_type_validation_kwargs():
    with given:
        actual_value = False
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.any(schema.none), actual_value, path=path)

    with then:
        assert result.get_errors() == [
            SchemaMismatchValidationError(path, actual_value, (schema.none,))
        ]



from typing import Any, List

import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import (
    ExtraElementValidationError,
    LengthValidationError,
    MaxLengthValidationError,
    MinLengthValidationError,
    MissingElementValidationError,
    TypeValidationError,
)


@pytest.mark.parametrize("value", [
    [],
    [1],
    [1, 2],
    [42, 3.14, "banana"],
])
def test_list_type_validation(value: List[Any]):
    with when:
        result = validate(schema.list, value)

    with then:
        assert result.get_errors() == []


def test_list_type_validation_error():
    with given:
        value = {}

    with when:
        result = validate(schema.list, value)

    with then:
        assert result.get_errors() == [
            TypeValidationError(PathHolder(), value, list),
        ]


def test_list_no_elements_validation():
    with when:
        result = validate(schema.list([]), [])

    with then:
        assert result.get_errors() == []


def test_list_homogeneous_elements_validation():
    with given:
        value = [1, 2]

    with when:
        result = validate(schema.list([
            schema.int(1),
            schema.int(2),
        ]), value)

    with then:
        assert result.get_errors() == []


def test_list_heterogeneous_elements_validation():
    with given:
        value = [42, 3.14, "banana"]

    with when:
        result = validate(schema.list([
            schema.int(42),
            schema.float(3.14),
            schema.str("banana"),
        ]), value)

    with then:
        assert result.get_errors() == []


def test_list_more_element_validation_error():
    with given:
        value = [1]

    with when:
        result = validate(schema.list([]), value)

    with then:
        assert result.get_errors() == [
            ExtraElementValidationError(PathHolder(), actual_value=value, index=0),
        ]


def test_list_more_elements_validation_error():
    with given:
        value = [1, 2, 3]

    with when:
        result = validate(schema.list([schema.int]), value)

    with then:
        assert result.get_errors() == [
            ExtraElementValidationError(PathHolder(), actual_value=value, index=1),
            ExtraElementValidationError(PathHolder(), actual_value=value, index=2),
        ]


def test_list_less_elements_validation_error():
    with given:
        value = [1, 2]

    with when:
        result = validate(schema.list([schema.int, schema.int, schema.int]), value)

    with then:
        assert result.get_errors() == [
            MissingElementValidationError(PathHolder(), actual_value=value, index=2),
        ]


def test_list_len_validation():
    with given:
        value = [1, 2]

    with when:
        result = validate(schema.list.len(2), value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize("value", [
    [1],
    [1, 2, 3],
])
def test_list_len_validation_error(value: List[Any]):
    with given:
        length = 2

    with when:
        result = validate(schema.list.len(length), value)

    with then:
        assert result.get_errors() == [
            LengthValidationError(PathHolder(), value, length)
        ]


@pytest.mark.parametrize("value", [
    [1, 2],
    [1, 2, 3],
])
def test_list_min_len_validation(value: List[Any]):
    with when:
        result = validate(schema.list.len(2, ...), value)

    with then:
        assert result.get_errors() == []


def test_list_min_len_validation_error():
    with given:
        value = [1]
        min_length = 2

    with when:
        result = validate(schema.list.len(min_length, ...), value)

    with then:
        assert result.get_errors() == [
            MinLengthValidationError(PathHolder(), value, min_length)
        ]


@pytest.mark.parametrize("value", [
    [1, 2],
    [1],
])
def test_list_max_len_validation(value: List[Any]):
    with when:
        result = validate(schema.list.len(..., 2), value)

    with then:
        assert result.get_errors() == []


def test_list_max_len_validation_error():
    with given:
        value = [1, 2, 3]
        max_length = 2

    with when:
        result = validate(schema.list.len(..., max_length), value)

    with then:
        assert result.get_errors() == [
            MaxLengthValidationError(PathHolder(), value, max_length)
        ]


@pytest.mark.parametrize(("min_length", "max_length"), [
    (2, 2),
    (1, 3),
])
def test_list_min_max_len_validation(min_length: int, max_length: int):
    with given:
        value = [1, 2]

    with when:
        result = validate(schema.list.len(min_length, max_length), value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize(("min_length", "max_length"), [
    (1, 1),
    (3, 3),
])
def test_list_min_max_len_validation_error(min_length: int, max_length: int):
    with given:
        value = [1, 2]

    with when:
        result = validate(schema.list.len(min_length, max_length), value)

    with then:
        assert len(result.get_errors()) == 1



import uuid

import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder, _

from valera import Formatter
from valera.errors import (
    AlphabetValidationError,
    ExtraElementValidationError,
    ExtraKeyValidationError,
    InvalidUUIDVersionValidationError,
    LengthValidationError,
    MaxLengthValidationError,
    MaxValueValidationError,
    MinLengthValidationError,
    MinValueValidationError,
    MissingElementValidationError,
    MissingKeyValidationError,
    RegexValidationError,
    SchemaMismatchValidationError,
    SubstrValidationError,
    TypeValidationError,
    ValueValidationError,
)


@pytest.fixture()
def formatter() -> Formatter:
    return Formatter()


def test_formatter_default_root():
    with when:
        formatter = Formatter()

    with then:
        assert formatter.root == "_"


def test_formatter_custom_root():
    with when:
        formatter = Formatter("#")

    with then:
        assert formatter.root == "#"


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value 'banana' must be <class 'int'>, but <class 'str'> given"),
    (_["id"], "Value 'banana' at _['id'] must be <class 'int'>, but <class 'str'> given")
])
def test_format_type_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = TypeValidationError(path, actual_value="banana", expected_type=int)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must be equal to 'banana', but 'orange' given"),
    (_["id"], "Value <class 'str'> at _['id'] must be equal to 'banana', but 'orange' given")
])
def test_format_value_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = ValueValidationError(path, actual_value="orange", expected_value="banana")

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'int'> must be greater than or equal to 1, but 0 given"),
    (_["id"], "Value <class 'int'> at _['id'] must be greater than or equal to 1, but 0 given")
])
def test_format_min_value_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = MinValueValidationError(path, actual_value=0, min_value=1)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'int'> must be less than or equal to 0, but 1 given"),
    (_["id"], "Value <class 'int'> at _['id'] must be less than or equal to 0, but 1 given")
])
def test_format_max_value_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = MaxValueValidationError(path, actual_value=1, max_value=0)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must have exactly 1 element, but it has 2 elements"),
    (_["id"], "Value <class 'str'> at _['id'] must have exactly 1 element, but it has 2 elements")
])
def test_format_length_one_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = LengthValidationError(path, actual_value="ab", length=1)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must have exactly 2 elements, but it has 1 element"),
    (_["id"], "Value <class 'str'> at _['id'] must have exactly 2 elements, but it has 1 element")
])
def test_format_length_many_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = LengthValidationError(path, actual_value="a", length=2)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must have at least 1 element, but it has 0 elements"),
    (_["id"], "Value <class 'str'> at _['id'] must have at least 1 element, but it has 0 elements")
])
def test_format_min_length_one_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = MinLengthValidationError(path, actual_value="", min_length=1)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must have at least 3 elements, but it has 1 element"),
    (_["id"], "Value <class 'str'> at _['id'] must have at least 3 elements, but it has 1 element")
])
def test_format_min_length_many_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = MinLengthValidationError(path, actual_value="a", min_length=3)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must have at most 1 element, but it has 2 elements"),
    (_["id"], "Value <class 'str'> at _['id'] must have at most 1 element, but it has 2 elements")
])
def test_format_max_length_one_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = MaxLengthValidationError(path, actual_value="ab", max_length=1)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must have at most 0 elements, but it has 1 element"),
    (_["id"], "Value <class 'str'> at _['id'] must have at most 0 elements, but it has 1 element")
])
def test_format_max_length_many_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = MaxLengthValidationError(path, actual_value="a", max_length=0)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must contain only '0123456789', but 'banana' given"),
    (_["id"], "Value <class 'str'> at _['id'] must contain only '0123456789', but 'banana' given")
])
def test_format_alphabet_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = AlphabetValidationError(path, actual_value="banana", alphabet="0123456789")

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must contain 'banana', but 'ananab' given"),
    (_["id"], "Value <class 'str'> at _['id'] must contain 'banana', but 'ananab' given")
])
def test_format_substr_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = SubstrValidationError(path, actual_value="ananab", substr="banana")

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'str'> must match pattern '[0-9]+', but 'banana' given"),
    (_["id"], "Value <class 'str'> at _['id'] must match pattern '[0-9]+', but 'banana' given")
])
def test_format_regex_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = RegexValidationError(path, actual_value="banana", pattern="[0-9]+")

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Element _[1] does not exist"),
    (_["id"], "Element _['id'][1] does not exist")
])
def test_format_missing_element_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = MissingElementValidationError(path, actual_value=["a"], index=1)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value contains extra element at index 1"),
    (_["id"], "Value at _['id'] contains extra element at index 1")
])
def test_format_extra_element_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = ExtraElementValidationError(path, actual_value=["a", "b"], index=1)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Key _['missing_key'] does not exist"),
    (_["id"], "Key _['id']['missing_key'] does not exist")
])
def test_format_missing_key_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = MissingKeyValidationError(path, actual_value={}, missing_key="missing_key")

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value contains extra key 'extra_key'"),
    (_["id"], "Value at _['id'] contains extra key 'extra_key'")
])
def test_format_extra_key_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        value = {"extra_key": "value"}
        error = ExtraKeyValidationError(path, actual_value=value, extra_key="extra_key")

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'int'> must match any of (schema.str, schema.none), but 42 given"),
    (_["id"], "Value <class 'int'> at _['id'] must match any of (schema.str, schema.none), "
              "but 42 given"),
])
def test_format_schema_missmatch_error(path: PathHolder, formatted: str, *, formatter: Formatter):
    with given:
        error = SchemaMismatchValidationError(path, actual_value=42,
                                              expected_schemas=(schema.str, schema.none))

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted


@pytest.mark.parametrize(("path", "formatted"), [
    (_, "Value <class 'uuid.UUID'> must be a UUID version 4, "
        "but UUID('886313e1-3b8a-5372-9b90-0c9aee199e5d') version 5 given"),
    (_["id"], "Value <class 'uuid.UUID'> at _['id'] must be a UUID version 4, "
              "but UUID('886313e1-3b8a-5372-9b90-0c9aee199e5d') version 5 given"),
])
def test_format_invalid_uuid_version_error(path: PathHolder, formatted: str, *,
                                           formatter: Formatter):
    with given:
        uuid5 = uuid.UUID("886313e1-3b8a-5372-9b90-0c9aee199e5d", version=5)
        error = InvalidUUIDVersionValidationError(path, actual_value=uuid5, actual_version=5,
                                                  expected_version=4)

    with when:
        res = error.format(formatter)

    with then:
        assert res == formatted



from typing import Any

from district42 import GenericSchema
from district42.types import Schema

from ._abstract_formatter import AbstractFormatter
from ._formatter import Formatter
from ._validation_result import ValidationResult
from ._validator import Validator
from ._version import version

__version__ = version
__all__ = ("validate", "validate_or_fail", "eq", "Validator", "ValidationResult",
           "ValidationException", "Formatter", "AbstractFormatter",)


_validator = Validator()
_formatter = Formatter()


def validate(schema: GenericSchema, value: Any, **kwargs: Any) -> ValidationResult:
    return schema.__accept__(_validator, value=value, **kwargs)


class ValidationException(Exception):
    pass


def validate_or_fail(schema: GenericSchema, value: Any, **kwargs: Any) -> bool:
    result = validate(schema, value, **kwargs)
    errors = [e.format(_formatter) for e in result.get_errors()]
    if len(errors) == 0:
        return True
    message = "\n - " + "\n - ".join(errors)
    raise ValidationException(message)


def eq(schema: GenericSchema, value: Any) -> bool:
    if isinstance(value, Schema):
        return isinstance(value, schema.__class__) and (schema.props == value.props)
    return not validate(schema, value=value).has_errors()


Schema.__override__(Schema.__eq__.__name__, eq)



from typing import Any, List

import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import (
    LengthValidationError,
    MaxLengthValidationError,
    MinLengthValidationError,
    TypeValidationError,
    ValueValidationError,
)


@pytest.mark.parametrize("value", [
    [],
    [1],
    [1, 2],
])
def test_list_of_elements_validation(value: Any):
    with when:
        result = validate(schema.list(schema.int), value)

    with then:
        assert result.get_errors() == []


def test_list_of_type_element_validation_error():
    with given:
        value = 3.14

    with when:
        result = validate(schema.list(schema.int), [value])

    with then:
        path = PathHolder()[0]
        assert result.get_errors() == [TypeValidationError(path, value, int)]


def test_list_of_type_elements_validation_error():
    with given:
        value = 3.14

    with when:
        result = validate(schema.list(schema.int), [42, value])

    with then:
        path = PathHolder()[1]
        assert result.get_errors() == [TypeValidationError(path, value, int)]


def test_list_of_type_element_value_validation_error():
    with given:
        expected_value = 42
        actual_value = 43

    with when:
        result = validate(schema.list(schema.int(expected_value)), [actual_value])

    with then:
        path = PathHolder()[0]
        assert result.get_errors() == [ValueValidationError(path, actual_value, expected_value)]


def test_list_of_type_elements_value_validation_error():
    with given:
        expected_value = 42
        actual_value = 43

    with when:
        result = validate(schema.list(schema.int(expected_value)), [expected_value, actual_value])

    with then:
        path = PathHolder()[1]
        assert result.get_errors() == [ValueValidationError(path, actual_value, expected_value)]


def test_list_of_len_validation():
    with given:
        value = [1, 2]

    with when:
        result = validate(schema.list(schema.int).len(2), value)

    with given:
        assert result.get_errors() == []


@pytest.mark.parametrize("value", [
    [1],
    [1, 2, 3],
])
def test_list_of_len_validation_error(value: List[Any]):
    with given:
        length = 2

    with when:
        result = validate(schema.list(schema.int).len(length), value)

    with then:
        assert result.get_errors() == [
            LengthValidationError(PathHolder(), value, length)
        ]


@pytest.mark.parametrize("value", [
    [1, 2],
    [1, 2, 3],
])
def test_list_of_min_len_validation(value: List[Any]):
    with when:
        result = validate(schema.list(schema.int).len(2, ...), value)

    with given:
        assert result.get_errors() == []


def test_list_of_min_len_validation_error():
    with given:
        value = [1]
        min_length = 2

    with when:
        result = validate(schema.list(schema.int).len(min_length, ...), value)

    with then:
        assert result.get_errors() == [
            MinLengthValidationError(PathHolder(), value, min_length)
        ]


@pytest.mark.parametrize("value", [
    [1, 2],
    [1],
])
def test_list_of_max_len_validation(value: List[Any]):
    with when:
        result = validate(schema.list(schema.int).len(..., 2), value)

    with given:
        assert result.get_errors() == []


def test_list_of_max_len_validation_error():
    with given:
        value = [1, 2, 3]
        max_length = 2

    with when:
        result = validate(schema.list(schema.int).len(..., max_length), value)

    with then:
        assert result.get_errors() == [
            MaxLengthValidationError(PathHolder(), value, max_length)
        ]


@pytest.mark.parametrize(("min_length", "max_length"), [
    (2, 2),
    (1, 3),
])
def test_list_of_min_max_len_validation(min_length: int, max_length: int):
    with given:
        value = [1, 2]

    with when:
        result = validate(schema.list(schema.int).len(min_length, max_length), value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize(("min_length", "max_length"), [
    (1, 1),
    (3, 3),
])
def test_list_of_min_max_len_validation_error(min_length: int, max_length: int):
    with given:
        value = [1, 2]

    with when:
        result = validate(schema.list(schema.int).len(min_length, max_length), value)

    with then:
        assert len(result.get_errors()) == 1


def test_list_of_type_validation_kwargs():
    with given:
        value = 3.14
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.list(schema.int), [42, value], path=path)

    with then:
        assert result.get_errors() == [
            TypeValidationError(path[1], value, int)
        ]



import uuid
from uuid import UUID, uuid4, uuid5

from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import (
    InvalidUUIDVersionValidationError,
    TypeValidationError,
    ValueValidationError,
)


def test_uuid4_type_validation():
    with when:
        result = validate(schema.uuid4, uuid4())

    with then:
        assert result.get_errors() == []


def test_uuid4_type_validation_error():
    with given:
        value = str(uuid4())

    with when:
        result = validate(schema.uuid4, value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder(), value, UUID)]


def test_uuid4_value_validation():
    with given:
        value = uuid4()

    with when:
        result = validate(schema.uuid4(value), value)

    with then:
        assert result.get_errors() == []


def test_uuid4_value_validation_error():
    with given:
        expected_value = uuid4()
        actual_value = uuid4()

    with when:
        result = validate(schema.uuid4(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value)
        ]


def test_uuid4_version_validation_error():
    with given:
        expected_value = uuid4()
        actual_value = uuid5(uuid.NAMESPACE_DNS, "python.org")

    with when:
        result = validate(schema.uuid4(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            InvalidUUIDVersionValidationError(PathHolder(), actual_value,
                                              actual_value.version, expected_value.version),
        ]



from datetime import datetime, timedelta

from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import TypeValidationError, ValueValidationError


def test_datetime_type_validation():
    with given:
        value = datetime.now()

    with when:
        result = validate(schema.datetime, value)

    with then:
        assert result.get_errors() == []


def test_datetime_type_validation_error():
    with given:
        value = str(datetime.now())

    with when:
        result = validate(schema.datetime, value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder(), value, datetime)]


def test_datetime_value_validation():
    with given:
        value = datetime.now()

    with when:
        result = validate(schema.datetime(value), value)

    with then:
        assert result.get_errors() == []


def test_datetime_value_validation_error():
    with given:
        expected_value = datetime.now()
        actual_value = datetime.now() + timedelta(days=1)

    with when:
        result = validate(schema.datetime(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value)
        ]


def test_datetime_type_validation_kwargs():
    with given:
        expected_value = datetime.now()
        actual_value = datetime.now() + timedelta(days=1)
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.datetime(expected_value), actual_value, path=path)

    with then:
        assert result.get_errors() == [
            ValueValidationError(path, actual_value, expected_value)
        ]



import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import TypeValidationError, ValueValidationError


@pytest.mark.parametrize("value", [True, False])
def test_bool_type_validation(value: bool):
    with when:
        result = validate(schema.bool, value)

    with then:
        assert result.get_errors() == []


def test_bool_type_validation_error():
    with given:
        value = "True"

    with when:
        result = validate(schema.bool, value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder(), value, bool)]


@pytest.mark.parametrize("value", [True, False])
def test_bool_value_validation(value: bool):
    with when:
        result = validate(schema.bool(value), value)

    with then:
        assert result.get_errors() == []


def test_bool_value_validation_error():
    with given:
        expected_value = True
        actual_value = False

    with when:
        result = validate(schema.bool(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value)
        ]


def test_bool_type_validation_kwargs():
    with given:
        expected_value = True
        actual_value = False
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.bool(expected_value), actual_value, path=path)

    with then:
        assert result.get_errors() == [
            ValueValidationError(path, actual_value, expected_value)
        ]



import re
from copy import deepcopy
from datetime import datetime
from math import isclose
from typing import Any, Callable, List, Optional, Type, cast
from uuid import UUID

from district42 import GenericSchema, SchemaVisitor
from district42.types import (
    AnySchema,
    BoolSchema,
    BytesSchema,
    ConstSchema,
    DateTimeSchema,
    DictSchema,
    FloatSchema,
    GenericTypeAliasSchema,
    IntSchema,
    ListSchema,
    NoneSchema,
    StrSchema,
    TypeAliasPropsType,
    UUID4Schema,
)
from district42.utils import is_ellipsis
from niltype import Nil, Nilable
from th import PathHolder

from ._validation_result import ValidationResult
from .errors import (
    AlphabetValidationError,
    ExtraElementValidationError,
    ExtraKeyValidationError,
    InvalidUUIDVersionValidationError,
    LengthValidationError,
    MaxLengthValidationError,
    MaxValueValidationError,
    MinLengthValidationError,
    MinValueValidationError,
    MissingElementValidationError,
    MissingKeyValidationError,
    RegexValidationError,
    SchemaMismatchValidationError,
    SubstrValidationError,
    TypeValidationError,
    ValidationError,
    ValueValidationError,
)

__all__ = ("Validator",)


class Validator(SchemaVisitor[ValidationResult]):
    def __init__(self, *,
                 validation_result_factory: Callable[[], ValidationResult] = ValidationResult,
                 path_holder_factory: Callable[[], PathHolder] = PathHolder) -> None:
        self._validation_result_factory = validation_result_factory
        self._path_holder_factory = path_holder_factory

    def _validate_type(self, path: PathHolder, value: Any,
                       expected_type: Type[Any]) -> Optional[ValidationError]:
        if not isinstance(value, expected_type):
            return TypeValidationError(path, value, expected_type)
        return None

    def _validate_value(self, path: PathHolder, value: Any,
                        expected_val: Any) -> Optional[ValidationError]:
        if value != expected_val:
            return ValueValidationError(path, value, expected_val)
        return None

    def _validate_elements(self,
                           path: PathHolder,
                           value: List[Any],
                           elements: List[GenericSchema],
                           start: int = 0,
                           **kwargs: Any) -> List[ValidationError]:
        errors: List[ValidationError] = []
        for index, element_schema in enumerate(elements):
            real_index = start + index
            try:
                val = value[real_index]
            except IndexError:
                errors.append(MissingElementValidationError(path, value, real_index))
                break
            else:
                nested_path = deepcopy(path)[real_index]
                res = element_schema.__accept__(self, value=val, path=nested_path, **kwargs)
                errors += res.get_errors()
        return errors

    def visit_none(self, schema: NoneSchema, *,
                   value: Any = Nil, path: Nilable[PathHolder] = Nil,
                   **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, type(None)):
            return result.add_error(error)

        return result

    def visit_bool(self, schema: BoolSchema, *,
                   value: Any = Nil, path: Nilable[PathHolder] = Nil,
                   **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, bool):
            return result.add_error(error)

        if schema.props.value is not Nil:
            if error := self._validate_value(path, value, schema.props.value):
                return result.add_error(error)

        return result

    def visit_int(self, schema: IntSchema, *,
                  value: Any = Nil, path: Nilable[PathHolder] = Nil,
                  **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, int):
            return result.add_error(error)

        if schema.props.value is not Nil:
            if error := self._validate_value(path, value, schema.props.value):
                return result.add_error(error)

        if schema.props.min is not Nil:
            if value < schema.props.min:
                result.add_error(MinValueValidationError(path, value, schema.props.min))

        if schema.props.max is not Nil:
            if value > schema.props.max:
                result.add_error(MaxValueValidationError(path, value, schema.props.max))

        return result

    def visit_float(self, schema: FloatSchema, *,
                    value: Any = Nil, path: Nilable[PathHolder] = Nil,
                    **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, float):
            return result.add_error(error)

        if schema.props.value is not Nil:
            if schema.props.precision is Nil:
                if not isclose(value, schema.props.value):
                    return result.add_error(ValueValidationError(path, value, schema.props.value))
            else:
                scale_factor = 10 ** schema.props.precision
                scaled_actual = round(value * scale_factor)
                scaled_expected = round(schema.props.value * scale_factor)
                if not isclose(scaled_expected, scaled_actual, rel_tol=0, abs_tol=0):
                    return result.add_error(ValueValidationError(path, value, schema.props.value))

        if schema.props.min is not Nil:
            if value < schema.props.min:
                result.add_error(MinValueValidationError(path, value, schema.props.min))

        if schema.props.max is not Nil:
            if value > schema.props.max:
                result.add_error(MaxValueValidationError(path, value, schema.props.max))

        return result

    def visit_str(self, schema: StrSchema, *,
                  value: Any = Nil, path: Nilable[PathHolder] = Nil,
                  **kwargs: Any) -> Any:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, str):
            return result.add_error(error)

        if schema.props.value is not Nil:
            if error := self._validate_value(path, value, schema.props.value):
                return result.add_error(error)

        if schema.props.pattern is not Nil:
            match_object = re.search(schema.props.pattern, value)
            if match_object is None:
                error = RegexValidationError(path, value, schema.props.pattern)
                return result.add_error(error)

        if schema.props.len is not Nil:
            if len(value) != schema.props.len:
                result.add_error(LengthValidationError(path, value, schema.props.len))
        if schema.props.min_len is not Nil:
            if len(value) < schema.props.min_len:
                result.add_error(MinLengthValidationError(path, value, schema.props.min_len))
        if schema.props.max_len is not Nil:
            if len(value) > schema.props.max_len:
                result.add_error(MaxLengthValidationError(path, value, schema.props.max_len))

        if schema.props.substr is not Nil:
            if schema.props.substr not in value:
                result.add_error(SubstrValidationError(path, value, schema.props.substr))

        if schema.props.alphabet is not Nil:
            alphabet = set(schema.props.alphabet)
            for letter in value:
                if letter not in alphabet:
                    return result.add_error(
                        AlphabetValidationError(PathHolder(), value, schema.props.alphabet))

        return result

    def visit_list(self, schema: ListSchema, *,
                   value: Any = Nil, path: Nilable[PathHolder] = Nil,
                   **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, list):
            return result.add_error(error)

        if schema.props.len is not Nil:
            if len(value) != schema.props.len:
                return result.add_error(LengthValidationError(path, value, schema.props.len))
        if schema.props.min_len is not Nil:
            if len(value) < schema.props.min_len:
                return result.add_error(
                    MinLengthValidationError(path, value, schema.props.min_len))
        if schema.props.max_len is not Nil:
            if len(value) > schema.props.max_len:
                return result.add_error(
                    MaxLengthValidationError(path, value, schema.props.max_len))

        if (schema.props.type is Nil) and (schema.props.elements is Nil):
            return result

        if schema.props.type is not Nil:
            type_schema = schema.props.type
            for index, elem in enumerate(value):
                nested_path = deepcopy(path)[index]
                res = type_schema.__accept__(self, value=elem, path=nested_path, **kwargs)
                result.add_errors(res.get_errors())
            return result

        elements = cast(List[GenericSchema], schema.props.elements)

        # body
        if (len(elements) > 2) and is_ellipsis(elements[0]) and is_ellipsis(elements[-1]):
            if len(value) == 0:
                errors = self._validate_elements(path, value, elements[1:-1], **kwargs)
                return result.add_errors(errors)
            all_errors = []
            for index, val in enumerate(value):
                errors = self._validate_elements(path, value, elements[1:-1], index, **kwargs)
                all_errors.append(errors)
            all_errors.sort(key=len)
            return result.add_errors(all_errors[0])

        # head
        if (len(elements) >= 2) and is_ellipsis(elements[-1]):
            errors = self._validate_elements(path, value, elements[:-1], **kwargs)
            return result.add_errors(errors)

        # tail
        if (len(elements) >= 1) and is_ellipsis(elements[0]):
            elements = elements[1:]
            start = max(0, len(value) - len(elements))
            errors = self._validate_elements(path, value, elements, start, **kwargs)
            return result.add_errors(errors)

        errors = self._validate_elements(path, value, elements, **kwargs)
        result.add_errors(errors)
        if len(value) > len(elements):
            for index in range(len(elements), len(value)):
                result.add_error(ExtraElementValidationError(path, value, index))

        return result

    def visit_dict(self, schema: DictSchema, *,
                   value: Any = Nil, path: Nilable[PathHolder] = Nil,
                   **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, dict):
            return result.add_error(error)

        if schema.props.keys is Nil:
            return result

        for key, (val, is_optional) in schema.props.keys.items():
            if is_ellipsis(key):
                continue
            if key in value:
                nested_path = deepcopy(path)[key]
                res = val.__accept__(self, value=value[key], path=nested_path, **kwargs)
                result.add_errors(res.get_errors())
            else:
                if not is_optional:
                    result.add_error(MissingKeyValidationError(path, value, key))

        if (... not in schema.props.keys) and (len(schema.props.keys) != len(value)):
            for key, val in value.items():
                if key not in schema.props.keys:
                    result.add_error(ExtraKeyValidationError(path, value, key))

        return result

    def visit_any(self, schema: AnySchema, *,
                  value: Any = Nil, path: Nilable[PathHolder] = Nil,
                  **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if schema.props.types is Nil:
            return result

        for sch_type in schema.props.types:
            res = sch_type.__accept__(self, path=path, value=value, **kwargs)
            if not res.has_errors():
                return result

        result.add_error(SchemaMismatchValidationError(path, value, schema.props.types))
        return result

    def visit_const(self, schema: ConstSchema, *,
                    value: Any = Nil, path: Nilable[PathHolder] = Nil,
                    **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if schema.props.value is not Nil:
            if error := self._validate_value(path, value, schema.props.value):
                return result.add_error(error)

        return result

    def visit_bytes(self, schema: BytesSchema, *,
                    value: Any = Nil, path: Nilable[PathHolder] = Nil,
                    **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, bytes):
            return result.add_error(error)

        if schema.props.value is not Nil:
            if error := self._validate_value(path, value, schema.props.value):
                return result.add_error(error)

        return result

    def visit_type_alias(self, schema: GenericTypeAliasSchema[TypeAliasPropsType], *,
                         value: Any = Nil, path: Nilable[PathHolder] = Nil,
                         **kwargs: Any) -> ValidationResult:
        return schema.props.type.__accept__(self, value=value, path=path, **kwargs)

    def visit_datetime(self, schema: DateTimeSchema, *,
                       value: Any = Nil, path: Nilable[PathHolder] = Nil,
                       **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, datetime):
            return result.add_error(error)

        if schema.props.value is not Nil:
            if error := self._validate_value(path, value, schema.props.value):
                return result.add_error(error)

        return result

    def visit_uuid4(self, schema: UUID4Schema, *,
                    value: Any = Nil, path: Nilable[PathHolder] = Nil,
                    **kwargs: Any) -> ValidationResult:
        result = self._validation_result_factory()
        if path is Nil:
            path = self._path_holder_factory()

        if error := self._validate_type(path, value, UUID):
            return result.add_error(error)

        if value.version != 4:
            return result.add_error(
                InvalidUUIDVersionValidationError(path, value, value.version, 4))

        if schema.props.value is not Nil:
            if error := self._validate_value(path, value, schema.props.value):
                return result.add_error(error)

        return result



from typing import Any, Dict

import pytest
from baby_steps import given, then, when
from district42 import optional, schema
from th import PathHolder

from valera import validate
from valera.errors import TypeValidationError


@pytest.mark.parametrize("value", [
    {"id": 1},
    {"id": 1, "name": "Bob"},
])
def test_dict_with_optional_key_validation(value: Dict[Any, Any]):
    with given:
        sch = schema.dict({
            "id": schema.int,
            optional("name"): schema.str,
        })

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_dict_with_optional_key_validation_error():
    with given:
        sch = schema.dict({
            "id": schema.int,
            optional("name"): schema.str,
        })

    with when:
        result = validate(sch, {
            "id": 1,
            "name": None
        })

    with then:
        assert result.get_errors() == [
            TypeValidationError(PathHolder()["name"], None, str)
        ]


def test_dict_relaxed_with_optional_key_validation_error():
    with given:
        sch = schema.dict({
            "id": schema.int,
            optional("name"): schema.str,
            ...: ...
        })

    with when:
        result = validate(sch, {
            "id": 1,
            "name": None
        })

    with then:
        assert result.get_errors() == [
            TypeValidationError(PathHolder()["name"], None, str)
        ]



from abc import ABC, abstractmethod
from typing import Any

from .errors import (
    AlphabetValidationError,
    ExtraElementValidationError,
    ExtraKeyValidationError,
    LengthValidationError,
    MaxLengthValidationError,
    MaxValueValidationError,
    MinLengthValidationError,
    MinValueValidationError,
    MissingElementValidationError,
    MissingKeyValidationError,
    RegexValidationError,
    SchemaMismatchValidationError,
    SubstrValidationError,
    TypeValidationError,
    ValueValidationError,
)

__all__ = ("AbstractFormatter",)


class AbstractFormatter(ABC):
    @abstractmethod
    def format_type_error(self, error: TypeValidationError) -> str:
        pass

    @abstractmethod
    def format_value_error(self, error: ValueValidationError) -> str:
        pass

    @abstractmethod
    def format_min_value_error(self, error: MinValueValidationError) -> str:
        pass

    @abstractmethod
    def format_max_value_error(self, error: MaxValueValidationError) -> str:
        pass

    @abstractmethod
    def format_length_error(self, error: LengthValidationError) -> str:
        pass

    @abstractmethod
    def format_min_length_error(self, error: MinLengthValidationError) -> str:
        pass

    @abstractmethod
    def format_max_length_error(self, error: MaxLengthValidationError) -> str:
        pass

    @abstractmethod
    def format_alphabet_error(self, error: AlphabetValidationError) -> str:
        pass

    @abstractmethod
    def format_substr_error(self, error: SubstrValidationError) -> str:
        pass

    @abstractmethod
    def format_regex_error(self, error: RegexValidationError) -> str:
        pass

    @abstractmethod
    def format_missing_element_error(self, error: MissingElementValidationError) -> str:
        pass

    @abstractmethod
    def format_extra_element_error(self, error: ExtraElementValidationError) -> str:
        pass

    @abstractmethod
    def format_missing_key_error(self, error: MissingKeyValidationError) -> str:
        pass

    @abstractmethod
    def format_extra_key_error(self, error: ExtraKeyValidationError) -> str:
        pass

    @abstractmethod
    def format_schema_missmatch_error(self, error: SchemaMismatchValidationError) -> str:
        pass

    def __getattr__(self, name: Any) -> Any:
        raise AttributeError(f"{self.__class__.__name__!r} object has no attribute {name!r}")

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if kwargs.get("extend", False) is not True:
            return
        parent = cls.__bases__[0]
        assert issubclass(parent, AbstractFormatter)
        for name, value in cls.__dict__.items():
            if callable(value) and not name.startswith("_"):
                setattr(parent, name, value)



from typing import Any, List

import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import MissingElementValidationError, ValueValidationError


@pytest.mark.parametrize("value", [
    [],
    [1],
    [1, 2],
])
def test_list_contains_validation(value: List[Any]):
    with given:
        sch = schema.list([...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize("value", [
    [1, 2],
    [1, 2, 3],
    [1, 2, 3, 4],
])
def test_list_contains_head_validation(value: List[Any]):
    with given:
        sch = schema.list([schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_list_contains_head_validation_incorrect_element_error():
    with given:
        value = [2]
        sch = schema.list([schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[0], actual_value=value[0], expected_value=1),
            MissingElementValidationError(PathHolder(), actual_value=value, index=1)
        ]


def test_list_contains_head_validation_missing_element_error():
    with given:
        value = [1]
        sch = schema.list([schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            MissingElementValidationError(PathHolder(), actual_value=value, index=1)
        ]


def test_list_contains_head_validation_error_extra_element():
    with given:
        value = [0, 1, 2]
        sch = schema.list([schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[0], actual_value=value[0], expected_value=1),
            ValueValidationError(PathHolder()[1], actual_value=value[1], expected_value=2),
        ]


@pytest.mark.parametrize("value", [
    [1, 2],
    [0, 1, 2],
    [-1, 0, 1, 2],
])
def test_list_contains_tail_validation(value: List[Any]):
    with given:
        sch = schema.list([..., schema.int(1), schema.int(2)])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_list_contains_tail_validation_incorrect_element_error():
    with given:
        value = [2]
        sch = schema.list([..., schema.int(1), schema.int(2)])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[0], actual_value=value[0], expected_value=1),
            MissingElementValidationError(PathHolder(), actual_value=value, index=1),
        ]


def test_list_contains_tail_validation_missing_element_error():
    with given:
        value = [1]
        sch = schema.list([..., schema.int(1), schema.int(2)])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            MissingElementValidationError(PathHolder(), actual_value=value, index=1),
        ]


def test_list_contains_tail_validation_extra_element_error():
    with given:
        value = [1, 2, 3]
        sch = schema.list([..., schema.int(1), schema.int(2)])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[1], actual_value=value[1], expected_value=1),
            ValueValidationError(PathHolder()[2], actual_value=value[2], expected_value=2),
        ]


def test_list_contains_validation_incorrect_tail_element_error():
    with given:
        value = [1, 3]
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[1], actual_value=3, expected_value=2),
        ]


@pytest.mark.parametrize("value", [
    [1, 2],
    [1, 2, 3],
    [0, 1, 2],
    [0, 1, 2, 3],
])
def test_list_contains_body_validation(value: List[Any]):
    with given:
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


def test_list_contains_validation_incorrect_head_element_error():
    with given:
        value = [3, 2]
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[0], actual_value=3, expected_value=1),
        ]


def test_list_contains_validation_extra_head_element_error():
    with given:
        value = [0, 1]
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            MissingElementValidationError(PathHolder(), actual_value=value, index=2)
        ]


def test_list_contains_validation_extra_tail_element_error():
    with given:
        value = [2, 3]
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[0], actual_value=2, expected_value=1),
            ValueValidationError(PathHolder()[1], actual_value=3, expected_value=2),
        ]


def test_list_contains_validation_extra_body_element_error():
    with given:
        value = [1, 0, 2]
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[1], actual_value=0, expected_value=2),
        ]


def test_list_contains_validation_missing_tail_element_error():
    with given:
        value = [1]
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            MissingElementValidationError(PathHolder(), actual_value=value, index=1),
        ]


def test_list_contains_validation_missing_head_element_error():
    with given:
        value = [2]
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder()[0], actual_value=2, expected_value=1),
            MissingElementValidationError(PathHolder(), actual_value=value, index=1),
        ]


def test_list_contains_validation_incorrect_order_error():
    with given:
        value = [2, 1]
        sch = schema.list([..., schema.int(1), schema.int(2), ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            MissingElementValidationError(PathHolder(), actual_value=value, index=2),
        ]


def test_list_contains_validation_no_elements_error():
    with given:
        value = []
        sch = schema.list([..., schema.int, ...])

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            MissingElementValidationError(PathHolder(), actual_value=value, index=0),
        ]



import pytest
from baby_steps import given, then, when
from district42 import schema
from th import PathHolder

from valera import validate
from valera.errors import (
    MaxValueValidationError,
    MinValueValidationError,
    TypeValidationError,
    ValueValidationError,
)


def test_float_type_validation():
    with when:
        result = validate(schema.float, 3.14)

    with then:
        assert result.get_errors() == []


def test_float_type_validation_error():
    with given:
        value = 42

    with when:
        result = validate(schema.float, value)

    with then:
        assert result.get_errors() == [TypeValidationError(PathHolder(), value, float)]


def test_float_value_validation():
    with given:
        value = 3.14

    with when:
        result = validate(schema.float(value), value)

    with then:
        assert result.get_errors() == []


def test_float_value_validation_error():
    with given:
        expected_value = 3.14
        actual_value = 3.15

    with when:
        result = validate(schema.float(expected_value), actual_value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), actual_value, expected_value)
        ]


@pytest.mark.parametrize("value", [3.15, 3.14])
def test_float_min_value_validation(value: float):
    with when:
        result = validate(schema.float.min(3.14), value)

    with then:
        assert result.get_errors() == []


def test_float_min_value_validation_error():
    with given:
        min_value = 3.14
        actual_value = 3.13

    with when:
        result = validate(schema.float.min(min_value), actual_value)

    with then:
        assert result.get_errors() == [
            MinValueValidationError(PathHolder(), actual_value, min_value)
        ]


@pytest.mark.parametrize("value", [3.14, 3.13])
def test_float_max_value_validation(value: float):
    with when:
        result = validate(schema.float.max(3.14), value)

    with then:
        assert result.get_errors() == []


def test_float_max_value_validation_error():
    with given:
        max_value = 3.14
        actual_value = 3.15

    with when:
        result = validate(schema.float.max(max_value), actual_value)

    with then:
        assert result.get_errors() == [
            MaxValueValidationError(PathHolder(), actual_value, max_value)
        ]


@pytest.mark.parametrize("value", [3.13, 3.14, 3.15])
def test_float_min_max_value_validation(value: float):
    with when:
        result = validate(schema.float.min(3.13).max(3.15), value)

    with then:
        assert result.get_errors() == []


def test_float_min_max_greater_value_validation_error():
    with given:
        min_value = 3.13
        max_value = 3.15
        actual_value = 3.16

    with when:
        result = validate(schema.float.min(min_value).max(max_value), actual_value)

    with then:
        assert result.get_errors() == [
            MaxValueValidationError(PathHolder(), actual_value, max_value)
        ]


def test_float_min_max_less_value_validation_error():
    with given:
        min_value = 3.13
        max_value = 3.15
        actual_value = 3.12

    with when:
        result = validate(schema.float.min(min_value).max(max_value), actual_value)

    with then:
        assert result.get_errors() == [
            MinValueValidationError(PathHolder(), actual_value, min_value)
        ]


def test_float_type_validation_kwargs():
    with given:
        expected_value = 3.14
        actual_value = 3.15
        path = PathHolder().items[0]["key"]

    with when:
        result = validate(schema.float(expected_value), actual_value, path=path)

    with then:
        assert result.get_errors() == [
            ValueValidationError(path, actual_value, expected_value)
        ]


@pytest.mark.parametrize("value", [-3.14, 3.0, 3.1, 3.14])
def test_float_precision_validation(value: float):
    with given:
        precision = 3

    with when:
        result = validate(schema.float.precision(precision), value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize("value", [9.00, 9.001, 9.004])
def test_float_pos_value_with_precision_validation(value: float):
    with given:
        sch = schema.float(9.00).precision(2)

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize("value", [-9.00, -9.001, -9.004])
def test_float_neg_value_with_precision_validation(value: float):
    with given:
        sch = schema.float(-9.00).precision(2)

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == []


@pytest.mark.parametrize("value", [9.005, 9.01, -9.00])
def test_float_pos_value_with_precision_validation_error(value: float):
    with given:
        sch = schema.float(9.00).precision(2)

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), value, 9.00)
        ]


@pytest.mark.parametrize("value", [9.005, 9.01, -9.00])
def test_float_neg_value_with_precision_validation_error(value: float):
    with given:
        sch = schema.float(9.00).precision(2)

    with when:
        result = validate(sch, value)

    with then:
        assert result.get_errors() == [
            ValueValidationError(PathHolder(), value, 9.00)
        ]



from baby_steps import given, then, when
from th import PathHolder

from valera import ValidationResult
from valera.errors import TypeValidationError, ValidationError


def make_error() -> ValidationError:
    return TypeValidationError(PathHolder(), "", int)


def test_validation_result():
    with when:
        result = ValidationResult()

    with then:
        assert result.get_errors() == []
        assert not result.has_errors()


def test_validation_result_with_errors():
    with given:
        errors = [make_error(), make_error()]

    with when:
        result = ValidationResult(errors)

    with then:
        assert result.get_errors() == errors
        assert result.has_errors()


def test_validation_result_add_error():
    with given:
        result = ValidationResult()
        error = make_error()

    with when:
        result = result.add_error(error)

    with then:
        assert result.get_errors() == [error]
        assert result.has_errors()


def test_validation_result_add_errors():
    with given:
        result = ValidationResult()
        errors = [make_error(), make_error()]

    with when:
        result = result.add_errors(errors)

    with then:
        assert result.get_errors() == errors
        assert result.has_errors()


def test_validation_result_repr():
    with given:
        result = ValidationResult()

    with when:
        res = repr(result)

    with then:
        assert res == "ValidationResult()"


def test_validation_result_with_errors_repr():
    with given:
        errors = [make_error(), make_error()]
        result = ValidationResult(errors)

    with when:
        res = repr(result)

    with then:
        assert res == f"ValidationResult({errors!r})"
